/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package vcl

import (
	"strconv"
	"strings"
	"text/template"
)

const ingTmplSrc = `vcl 4.1;

import std;
import directors;
import re2;
import dynamic;

include "bogo_backend.vcl";

{{- define "ProbeDef"}}
{{- range $name, $svc := .}}
{{- if $svc.Probe}}
{{with $svc.Probe}}
probe {{probeName $name}} {
{{- if ne .URL ""}}
	.url = "{{.URL}}";
{{- else if .Request}}
	.request =
{{- range $reqLine := .Request}}
		"{{$reqLine}}"
{{- end}}
		;
{{- end}}
{{- if .ExpResponse}}
	.expected_response = {{.ExpResponse}};
{{- end}}
{{- if .Timeout}}
	.timeout = {{.Timeout}};
{{- end}}
{{- if .Interval}}
	.interval = {{.Interval}};
{{- end}}
{{- if .Initial}}
	.initial = {{.Initial}};
{{- end}}
{{- if .Window}}
	.window = {{.Window}};
{{- end}}
{{- if .Threshold}}
	.threshold = {{.Threshold}};
{{- end}}
}
{{- end}}
{{- end}}
{{- end}}
{{- end}}
{{- template "ProbeDef" .IntSvcs -}}
{{- template "ProbeDef" .ExtSvcs}}

{{range $name, $svc := .IntSvcs -}}
{{range $addr := $svc.Addresses -}}
backend {{backendName $svc $addr}} {
	.host = "{{$addr.IP}}";
	.port = "{{$addr.Port}}";
{{- with $svc}}
{{- if .HostHeader}}
	.host_header = "{{.HostHeader}}";
{{- end}}
{{- if .ConnectTimeout}}
	.connect_timeout = {{.ConnectTimeout}};
{{- end}}
{{- if .FirstByteTimeout}}
	.first_byte_timeout = {{.FirstByteTimeout}};
{{- end}}
{{- if .BetweenBytesTimeout}}
	.between_bytes_timeout = {{.BetweenBytesTimeout}};
{{- end}}
{{- if .ProxyHeader}}
	.proxy_header = {{.ProxyHeader}};
{{- end}}
{{- if .MaxConnections}}
	.max_connections = {{.MaxConnections}};
{{- end}}
{{- if .Probe}}
	.probe = {{probeName $name}};
{{- end}}
{{- end}}
}
{{end -}}
{{end}}

sub vcl_init {
{{- if .Rules}}
	new vk8s_hosts = re2.set(anchor=both);
	{{- range $rule := .Rules}}
	vk8s_hosts.add("{{hostRegex $rule.Host}}(:\d+)?");
	{{- end}}
	vk8s_hosts.compile();
{{end}}

{{- range $name, $svc := .IntSvcs}}
	{{- $dirType := dirType $svc}}
	new {{dirName $svc}} = directors.{{$dirType}}();
	{{- range $addr := $svc.Addresses}}
	{{dirName $svc}}.add_backend({{backendName $svc $addr}}
		{{- if eq $dirType "random"}}
		, 1.0
		{{- end}}
		);
	{{- end}}
	{{- if eq $dirType "shard"}}
	{{- if $svc.Director.Warmup}}
	{{dirName $svc}}.set_warmup({{$svc.Director.Warmup}});
	{{- end}}
	{{- if $svc.Director.Rampup}}
	{{dirName $svc}}.set_rampup({{$svc.Director.Rampup}});
	{{- end}}
	{{dirName $svc}}.reconfigure();
	{{- end}}
{{end}}
{{- range $name, $svc := .ExtSvcs}}
	new {{resolverName $svc}} = dynamic.resolver();
	{{resolverName $svc}}.set_resolution_type(STUB);
	new {{dirName $svc}} = dynamic.director(
		ttl_from = dns,
		ttl = {{.DNSRetryDelay}},
		resolver = {{resolverName $svc}}.use()
		{{- with $svc}}
		{{- if .ExternalPort}}
		, port = "{{.ExternalPort}}"
		{{- end}}
		{{- if .HostHeader}}
		, host_header = "{{.HostHeader}}"
		{{- end}}
		{{- if .ConnectTimeout}}
		, connect_timeout = {{.ConnectTimeout}}
		{{- end}}
		{{- if .FirstByteTimeout}}
		, first_byte_timeout = {{.FirstByteTimeout}}
		{{- end}}
		{{- if .BetweenBytesTimeout}}
		, between_bytes_timeout = {{.BetweenBytesTimeout}}
		{{- end}}
		{{- if .DomainUsageTimeout}}
		, domain_usage_timeout = {{.DomainUsageTimeout}}
		{{- end}}
		{{- if .FirstLookupTimeout}}
		, first_lookup_timeout = {{.FirstLookupTimeout}}
		{{- end}}
		{{- if .ProxyHeader}}
		, proxy_header = {{.ProxyHeader}}
		{{- end}}
		{{- if .MaxConnections}}
		, max_connections = {{.MaxConnections}}
		{{- end}}
		{{- if .Probe}}
		, probe = {{probeName $name}}
		{{- end}}
		{{- end}}
	);
	{{- if .ResolverIdleTimeout}}
	{{resolverName $svc}}.set_idle_timeout({{.ResolverIdleTimeout}});
	{{- end }}
	{{- if .ResolverTimeout}}
	{{resolverName $svc}}.set_timeout({{.ResolverTimeout}});
	{{- end }}
	{{- if .MaxDNSQueries}}
	{{resolverName $svc}}.set_limit_outstanding_queries({{.MaxDNSQueries}});
	{{- end }}
	{{- if not .FollowDNSRedirects}}
	{{resolverName $svc}}.set_follow_redirects(REDIRECTS_DO_NOT_FOLLOW);
	{{- end }}
{{end}}
{{- range $rule := .Rules}}
	new {{urlMatcher $rule}} = re2.set(posix_syntax=true, anchor=start);
	{{- range $path, $svc := $rule.PathMap}}
	{{urlMatcher $rule}}.add("{{$path}}",
		{{- if $svc.ExternalName}}
		backend={{dirName $svc}}.backend("{{$svc.ExternalName}}"));
		{{- else}}
				backend={{dirName $svc}}.backend());
		{{- end}}
	{{- end}}
	{{urlMatcher $rule}}.compile();
{{end -}}
}

sub vcl_backend_fetch {
	set bereq.backend = vk8s_notfound;
{{- if .Rules}}
	if (vk8s_hosts.match(bereq.http.Host)) {
		{{- range $i, $rule := .Rules}}
		{{ifelse $i}} (vk8s_hosts.which(select=FIRST) == {{plusOne $i}}) {
			if ({{urlMatcher $rule}}.match(bereq.url)) {
				set bereq.backend = {{urlMatcher $rule}}.backend(select=FIRST);
			}
		}
		{{- end}}
	}
{{- end}}

	if (bereq.backend == vk8s_notfound) {
{{- if .DefaultService.Name}}
		set bereq.backend = {{dirName .DefaultService}}.backend();
{{- else}}
		return (error(404));
{{- end}}
	}
}
`

const ingTmplName = "vcl"

func dirType(svc Service) string {
	if svc.Director == nil {
		return RoundRobin.String()
	}
	return svc.Director.Type.String()
}

func backendName(svc Service, addr Address) string {
	n := ""
	if addr.PodNamespace != "" && addr.PodName != "" {
		n = addr.PodNamespace + "_" + addr.PodName
	} else {
		n = svc.Name + "_" + strings.Replace(addr.IP, ".", "_", -1)
	}
	return mangle(n + "_" + strconv.Itoa(int(addr.Port)))
}

func urlMatcher(rule Rule) string {
	if rule.Host == "" {
		return mangle("any_host")
	}
	return mangle(strings.Replace(rule.Host, ".", "_", -1) + "_url")
}

func hostRegex(host string) string {
	if host == "" {
		return "[^:]+"
	}
	return `\Q` + host + `\E`
}

var vclFuncs = template.FuncMap{
	"plusOne":   func(i int) int { return i + 1 },
	"dirType":   func(svc Service) string { return dirType(svc) },
	"hostRegex": func(host string) string { return hostRegex(host) },
	"probeName": func(name string) string {
		return mangle(name + "_probe")
	},
	"backendName": func(svc Service, addr Address) string {
		return backendName(svc, addr)
	},
	"dirName": func(svc Service) string {
		return mangle(svc.Name + "_director")
	},
	"resolverName": func(svc Service) string {
		return mangle(svc.Name + "_resolver")
	},
	"urlMatcher": func(rule Rule) string {
		return urlMatcher(rule)
	},
	"ifelse": func(i int) string {
		if i == 0 {
			return "if"
		}
		return "elsif"
	},
}

var ingressTmpl = template.Must(template.New(ingTmplName).
	Funcs(vclFuncs).Parse(ingTmplSrc))
